% THD model for DTU Class D course
% Flemming Nyboe (fls@ti.com) 2006-08-14

% **** HOW TO USE THIS FILE ********
% You set system parameters (for example PVDD, RDSon, RL ...), and calculate THD vs. Power
% for different settings. Some parameters will be fixed through the execution, others can vary from
% graph to graph on the same THD plot. You can choose which ones, in the "SYSTEM PARAMETERS"
% section, by setting some parameters to a fixed value, while letting others depend on the for-loop
% variable "run".

% ***** ASSUMPTIONS AND MECHANISMS NOT INCLUDED IN MODEL ********
% 1) Body diode I/V is temperature independent, but adjusted for fixed knee voltage w. RDSon
% 2) THD caused by PSU impedance is not modeled

% Plot setup
PlotColor='brgkmcy';
Legend1='';
Legend2='';
Legend3='';
Legend4='';
close all;
fprintf('--------------------------------------------\n');
  
% Plots on/off
PLOT_FETIV= 1;        % figure 1
PLOT_AREAS_COND= 1;   % figure 2
PLOT_AREAS_SWITCH= 1; % figure 3
PLOT_THD  = 1;        % figure 4

for run=0:0,    % Set the range for run here - start at zero. It runs out of plot colors above run=6
  
  % ****** SYSTEM PARAMETERS BEGIN ******* 
  RDSon=0.100;      % RDSon, equal for HS and LS
  Vknee=0.8;        % Source-Drain knee voltage for diode/channel conduction
  RL=4;             % BTL load resistance
  Fsw=384000;       % Switching frequency, Hz
  Lout=10e-6;       % OPF inductor value
  MaxM=0.95;        % Actual modulation index at Mod=1
  PVDD=30.0;        % Half bridge supply voltage
  DT=1e-12;         % Dead time (0 not allowed)
  Csw=300e-12;      % Half bridge switching capacitance (for both switches in parallel)
  nsA_LH=0e-9;      % Forced commutation equivalent switching delay (ns/A)
  DynrdB=100;       % Noise is added to THD+N plots according to this dynamic range in dB
  % ****** SYSTEM PARAMETERS END *********   
  
  % Fixed parameters
  Vt=0.0256;                    % Body diode Vt
  Is=Vt/RDSon*exp(-Vknee/Vt);   % Body Diode Is set for right Knee voltage
  nsA_HL=nsA_LH;

  % Show System parameters
  fprintf('PVDD                        : %2.1f V\n',PVDD);
  fprintf('Peak ripple current at idle : %2.3f A\n',PVDD/8/Fsw/Lout);
  fprintf('Dead time                   : %2.3f ns\n',1e9*DT);
  fprintf('Switch output capacitance   : %2.0f pF\n',1e12*Csw);  
  fprintf('I_LIM for linear commutation: %2.3f A\n',PVDD*Csw/DT);
  fprintf('Rload (BTL)                 : %2.1f ohms\n',RL);
  
  % ****** CALCULATION PARAMETERS (don't change) ******
  ModSteps=201;       % No. of Modulation index steps (no. of points on the Modulation-to-voltage transfer characteristic)
  FETSteps=51;        % No. of points in FET on-state V-I characteristic (IDS range is -PVDD/RL to PVDD/RL)
  OnCurrentSteps=31;  % No. of FET current data points during FET on-state

  % Basic vectors
  Mod=linspace(-MaxM,MaxM,ModSteps)'; % Modulation index, 0=idle
  Vspk=(1+Mod)/2*PVDD;                % Idealized speaker terminal voltage, corresponding to Deadtime=0 and RDSon=0
  Ispk=(Vspk-PVDD/2)/(RL/2);          % Idealized speaker terminal output current, corresponding to Deadtime=0 and RDSon=0
  PWMpw=(1+Mod)/2/Fsw;                % Idelaized PWM High pulse width (no dead time)
  ILrip=PWMpw.*(PVDD-Vspk)/Lout/2;    % Peak ripple current (= 1/2 peak-peak), always positive
  
  % Write progress
  fprintf('** Run no. %d: FET IV .. ',run);
  
  % FET on-state Drain-Source voltage drops
  Igen=linspace(-PVDD/RL,PVDD/RL,FETSteps);  % Range FET D-S currents (for full modulation range -1 to 1)
  VDSgen=VDSOn(Igen,RDSon,Is,Vt);
  
  fprintf('HS&LS on .. ');
  
  % From the above general relation between FET current and voltage drop: Igen and VDSgen (for both positive and
  % negative currents), the actual voltage drop integrals (PWM area losses) during the HS FET and LS FET on stages are
  % derived. Note that the FET current during the ON-state varies due to ripple current.
  VDSavg_HS=zeros(ModSteps,1);
  VDSavg_LS=zeros(ModSteps,1);
  for g=1:ModSteps,
    if (Ispk(g)-ILrip(g)) < Igen(1) fprintf('\nERROR: Current out of range (too low)'); end;
    if (Ispk(g)+ILrip(g)) > Igen(FETSteps) fprintf('\nERROR: Current out of range (too high)'); end;
    Ion=linspace(Ispk(g)-ILrip(g),Ispk(g)+ILrip(g),OnCurrentSteps); % FET IDS during on-state
    VDSavg_HS(g)=Mean(interp1(Igen,VDSgen,Ion,'spline'));
  end;
  
  % When the HS FET is on, ID_HS=IL (output inductor current).
  % When the LS FET is on, ID_LS=-IL
  % The LS FET average VDS during its on-state is the same as the HS FET drop, only for the opposite current
  % Since the vector Ispk is symmetrical around zero (zero is its middle element) we can find 
  % VDSavg_LS just by flipping VDSavgHS
  VDSavg_LS=flipud(VDSavg_HS);  
  
  % *** Plot FET on-state average voltage drops ***
  if PLOT_FETIV,
    figure(1);
    h=plot(Igen,VDSgen,PlotColor(run+1)); hold on;
    set(h,'LineWidth',2);
    Legend1=strvcat(Legend1,sprintf('R_{DS(ON)}=%6.3f',RDSon)); %%% PLOT LEGEND, RDSon 
  end;
  
  VDS_HS_HL=interp1(Igen,VDSgen,Ispk+ILrip,'spline');
  VDS_LS_LH=interp1(Igen,VDSgen,-(Ispk-ILrip),'spline');
  
  % Write progress
  fprintf('PWM Areas .. ');
  
  % PWM voltage areas
  PWM_A1=PWMpw.*PVDD;                      % Ideal modulation
  PWM_A2=(PWMpw-DT).*-VDSavg_HS;           % HS FET ON state D-S voltage drop (subtracts from area)
  PWM_A3=(1/Fsw-PWMpw-DT).*VDSavg_LS;      % LS FET ON state D-S voltage drop (adds to area)
  PWM_A4=Alh(Ispk-ILrip,PVDD,Csw,DT);      % Voltage integral during LH transition dead time
  PWM_A5=Ahl(Ispk+ILrip,PVDD,Csw,DT);      % Voltage integral during HL transition dead time  
  PWM_A6=PVDD*min(0,-nsA_LH*(Ispk-ILrip)); % Finite-speed forced commutation LH transition
  PWM_A7=PVDD*max(0,-nsA_HL*(Ispk+ILrip)); % Finite-speed forced commutation HL transition

  % PWM_A4=0; PWM_A5=0; PWM_A6=0; PWM_A7=0;  % OVERRIDE: ignore all switching errors
  
  %*** PLOT ERROR FUNCTIONS *** 
  if PLOT_AREAS_COND,
    figure(2);
    h=plot(Ispk,PWM_A2+PWM_A3 +RDSon*Ispk*(1/Fsw-2*DT),PlotColor(run+1));   % last term is added to emphasize nonlinearity
    set(h,'LineWidth',2); hold on;
    Legend2=strvcat(Legend2,sprintf('R_{DS(ON)}=%6.3f',RDSon)); %%% PLOT LEGEND, RDSon
  end;
  if PLOT_AREAS_SWITCH,
    figure(3);
    h=plot(Ispk,PWM_A4+PWM_A6,[PlotColor(run+1),'-']);
    set(h,'LineWidth',2); hold on;
    h=plot(Ispk,PWM_A5+PWM_A7,[PlotColor(run+1),':']);
    set(h,'LineWidth',2);
    Legend3=strvcat(Legend3,sprintf('nsA=%4.1fns/A DT=%5.2fns LH',nsA_LH*1e9,DT*1e9)); %%% PLOT LEGEND, Delay & Dead time
    Legend3=strvcat(Legend3,sprintf('nsA=%4.1fns/A DT=%5.2fns HL',nsA_LH*1e9,DT*1e9)); %%% PLOT LEGEND, Delay & Dead time
  end;
  
  lvlOUT=(PWM_A1+PWM_A2+PWM_A3+PWM_A4+PWM_A5+PWM_A6+PWM_A7)/PVDD*2*Fsw-1;
  
  % Write progress
  fprintf('THD .. ');
  
  % *** FFT setup ***
  Nfft=32;                    % set FFT size (number of harmonics included in THD+N will be Nfft/2-1
  t=(0:Nfft-1)/Nfft*2*pi;     % time vector for input signal
  
  %%% ******** OUTPUT THD *********
  THDse=zeros(121,1);
  THDbtl=zeros(121,1);
  AdB=zeros(121,1);
  Apwr=zeros(121,1);
  for m=1:121,
    AdB(m)=(1-m)/4;  % dB range is 0dB to -30dB in 0.25dB steps
    A=10^(AdB(m)/20);
    sineOUT=interp1(Mod,lvlOUT,MaxM*A*cos(t),'spline');    
    
    S=fft(sineOUT)/Nfft;  % spectrum
    G=abs(S);             % real spectrum
    P=2*G.^2;             % power spectrum, watts into 1 ohm load at PVDD=1V
    
    % Summing up power of fft bins. Note that P vector index Nfft/2 is bin # Nfft/2-1, so no aliased
    % power is included. All power levels are RMS watts into 1 ohm (BTL) at PVDD=1V, so a full scale
    % undistorted sine wave will have Ptot = 1/2 W.
    Ptot=sum(P(2:Nfft/2));          % index 2 it the fundamental (index 1 is DC)
    Pdist=sum(P(3:Nfft/2));         % all harmonics, 2nd and up 
    Podd=sum(P(4:2:Nfft/2));        % odd harmonics only, 3rd, 5th ...
    Pnoise=0.5*10^(-DynrdB/10);     % Noise power corresponding to dynamic range DynrdB (0dB = 0.5W)
    
    Apwr(m)=Ptot*PVDD^2/RL;                    % Actual speaker power including all harmonics (for plot x-asis values, alternative to AdB)
    THDse(m)=sqrt((Pdist+Pnoise)/Ptot)*100;    % SE mode THD+N - includes all harmonics
    THDbtl(m)=sqrt((Podd+Pnoise)/Ptot)*100;    % BTL mode THD+N - includes only odd harmonics, since the even will cancel
  end;
  
  % Write progress
  fprintf('Done.\n');
  
  %%% plot THD curve
  if PLOT_THD,
    figure(4);
    h=loglog(Apwr,THDbtl,PlotColor(run+1));  % THD vs. Power     (change figure settings below also)
    set(h,'LineWidth',2);
    hold on;
    refresh;
    
    Legend4=strvcat(Legend4,sprintf('R_{DS(ON)}=%6.3f Dead time=%5.2fns',RDSon,DT*1e9)); %%% PLOT LEGEND, RDSon & Dead time
    % Legend4=strvcat(Legend4,sprintf('Force delay=%4.1fns/A Dead time=%5.2fns',nsA_LH*1e9,DT*1e9)); %%% PLOT LEGEND, Delay & Dead time
  end;
end;

%%%% FIGURE FORMATTING

if PLOT_FETIV,
  figure(1);
  set(gca,'FontSize',12,'FontWeight','bold');
  title('FET IV characteristic')
  xlabel('FET ID')
  ylabel('FET VDS')
  grid on;
  legend(Legend1,2)  
end;

if PLOT_AREAS_COND,
  figure(2);
  set(gca,'FontSize',12,'FontWeight','bold');
  title('PWM Conduction Area errors')
  xlabel('Ispk')
  ylabel('Area V*s')
  grid on;
  legend(Legend2,2)
  ylim([-200e-9 200e-9]);
end;

if PLOT_AREAS_SWITCH,
  figure(3);
  set(gca,'FontSize',12,'FontWeight','bold');
  title('PWM Switching Area errors')
  xlabel('Ispk')
  ylabel('Area V*s')
  grid on;
  legend(Legend3,1)
  ylim([-200e-9 200e-9]);
end;

if PLOT_THD,
  figure(4);
  set(gca,'FontSize',12,'FontWeight','bold');
  title(sprintf('THD+N vs. Power. V_{DD}=%dV R_L=%d ohms',PVDD,RL))
  ylabel('THD+N %')
  grid on;
  axis([0.1 200 0.001 10])   
  xlabel('Output Power')
  set(gca,'YTickLabel',[0.001 0.01 0.1 1 10],'XTickLabel',[0.1 1 10 100]); 
  legend(Legend4,2)
end;

